package com.metasploit.meterpreter;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;

/**
 * A channel for a started {@link Process}.
 *
 * @author mihi
 */
public class ProcessChannel extends Channel {

    private final Process process;
    private final InputStream err;

    /**
     * Create a new process channel.
     *
     * @param meterpreter The meterpreter this channel should be assigned to.
     * @param process     Process of the channel
     */
    public ProcessChannel(Meterpreter meterpreter, Process process) {
        super(meterpreter, process.getInputStream(), process.getOutputStream());
        this.process = process;
        this.err = process.getErrorStream();
        new StderrThread(err, meterpreter.getErrorStream()).start();
    }

    /**
     * Read at least one byte, and up to maxLength bytes from this stream.
     * An empty string (0 length data) is returned if no data is available.
     *
     * @param maxLength The maximum number of bytes to read.
     * @return The bytes read, or <code>null</code> if the end of the stream has been reached.
     */
    public synchronized byte[] read(int maxLength) throws IOException, InterruptedException {
        if (closed)
            return null;
        if (active)
            throw new IllegalStateException("Cannot read; currently interacting with this channel");
        if (!waiting || (toRead != null && toRead.length == 0))
            return new byte[0];
        if (toRead == null)
            return null;
        return super.read(maxLength);
    }

    public void close() throws IOException {
        process.destroy();
        err.close();
        super.close();
    }

    class StderrThread extends Thread {
        private final InputStream inputStream;
        private final PrintStream outputStream;

        public StderrThread(InputStream inputStream, PrintStream outputStream) {
            this.inputStream = inputStream;
            this.outputStream = outputStream;
        }

        public void run() {
            try {
                byte[] buffer = new byte[1024*1024];
                int len;
                while ((len = inputStream.read(buffer)) != -1) {
                    if (len == 0)
                        continue;
                    byte[] data = new byte[len];
                    System.arraycopy(buffer, 0, data, 0, len);
                    handleInteract(data);
                }
            } catch (Throwable t) {
                t.printStackTrace(this.outputStream);
            }
        }
    }

}
